package ESPOPS::Object;

use strict;
use vars qw($VERSION @ISA $CONF $TABLE_DEF $Revision);

BEGIN {
	$Revision = sprintf "%d.%03d", q$Revision: 1.20 $ =~ /: (\d+)\.(\d+)/;
	$VERSION = '0.4';
	@ISA = qw(ESPOPS::DBI);
	$CONF = {
		ObjectAlias => {
			class			=> 'ESPOPS::Object',
			isa				=> \@ISA,
			field			=> [ qw/ id class / ],
			as_string_order => [ qw/ id class / ],
			base_table		=> 'Object',
			id_field		=> 'id',
			increment_field => 1,
			no_security		=> 1,
		},
	};
	$TABLE_DEF = <<SQL;
CREATE TABLE Object (
	id int(11) AUTO_INCREMENT PRIMARY KEY,
	class  char(255),
	tstamp timestamp(14) NOT NULL
)
SQL
}

use constant ROOT_OBJ_CLASS => __PACKAGE__;

use ESPOPS::DBI;
use SPOPS::ClassFactory;
use DBI qw( :sql_types );

sub new {
    my $class = shift;
    my $p     = shift;
# 	my $this = $class->SUPER::new($p);
	
	## Since SUPER::new($p) ignores keys in $p that are defined as fields
	## in the CONFIG of a parent object, we have to do the assigning of
	## these parameters manually (or fix SPOPS to handle this internally).
	my $this = $class->SUPER::new;
	
	foreach my $field ( @{$class->all_fields} ) {
	    $this->{$field} = $p->{$field}	if defined $p->{$field};
	}
	
	$this->{class} = ref($this);

	return $this;
}

sub pre_remove_action {
	my $this	= shift;
	my $p		= shift;
	my $class	= ref($this);
	
	## call the overridden pre_remove_action to handle rulesets
	$this->SUPER::pre_remove_action($p);

	## remove all corresponding rows in 'links_to' tables
	my ($table, $where);
	foreach $table ( values %{$this->CONFIG->{links_to}} ) {
		$where = $this->id_clause(undef, 'noqualify', $p);
		eval { $this->db_delete( {	table => $table,
									where => $where,
									db    => $p->{db},
									DEBUG => $p->{DEBUG} }) };
		if ( $@ ) {
			warn "Unable to remove links."; 
		}
	}
	
	## remove corresponding row in each parent table
	foreach my $c ( @{$class->isa_classes} ) {
		next if $c eq $class;
		$table = $c->table_name;
		$where = $c->id_clause($this->id, 'noqualify', $p);
		eval { $this->db_delete( {	table => $table,
									where => $where,
									db    => $p->{db},
									DEBUG => $p->{DEBUG} }) };
		if ( $@ ) {
			warn "Unable to remove row from parent table."; 
		}
	}

	return $this;
}

sub pre_save_action {
	my $this	= shift;
	my $p		= shift;
	my $class	= ref($this);

	## call the overridden pre_remove_action to handle rulesets
	$this->SUPER::pre_save_action($p);

	## save attributes in parent tables, top level table assigns the ID
	my $is_add = ( $p->{is_add} or ! $this->saved );
	PARENT: foreach my $parent_class ( @{$class->CONFIG->{isa}} ) {
		next unless $parent_class->isa(ROOT_OBJ_CLASS);
		
		## create a new parent object
		my $parent_obj = $parent_class->new;

		## copy all parent fields except id field
		foreach my $field ( @{$parent_obj->all_fields} ) {
			$parent_obj->{$field} = $this->{$field}	unless $field eq $this->id_field;
		}
		
		if (my $id = $this->id) {
			## assign id field if it's defined
			$parent_obj->{$parent_obj->id_field} = $this->{$this->id_field};
			if ($is_add) {
				## must be multiple inheritance and we hit something saved
				## on previous traversal back to the root object
				my $args = {	from	=> [ $parent_class->table_name],
								select	=> [ $parent_class->id_field],
								where	=> $parent_class->id_clause($id, undef, $p),
								return => 'single'
							};
				my $row = eval { $parent_class->db_select( $args ); };
				next PARENT if $row;
			} else {
				## we're doing an update
				$parent_obj->has_save;
				$parent_obj->has_change;
			}
		}
		
		## save the parent object
		$this->{$this->id_field} = $parent_obj->save( { %$p, is_add => $is_add } )->id;
		
		## copy all fields from parent back to the object (except class and id)
		## in case something was changed during the save operation (e.g. object
		## auto-saves a sub-object and updates the corresponding field, such as
		## Boat and Anchor)
		foreach my $field ( @{$parent_obj->all_fields} ) {
			$this->{$field} = $parent_obj->{$field}
				unless $field eq 'class' || $field eq $parent_obj->id_field;
		}
	}

	return $this;
}

sub fetch  {
	my ( $class, $id, $p ) = @_;
	$id ||= 0;
	$p  ||= {};
	SPOPS::DEBUG() && _wm( 2, SPOPS::DBI::DEBUG_FETCH,
				"Trying to fetch an item of $class with ID $id and params ", 
				join " // ",
				map { $_ . ' -> ' . ( defined( $p->{$_} ) ? $p->{$_} : '' ) }
					keys %{ $p } );
 
  	# No ID, no object

  	return undef   unless ( $id and $id !~ /^tmp/ );
  
  	# Security violations bubble up to caller  

  	my $level = $p->{security_level};
  	unless ( $p->{skip_security} ) {
    		$level ||= $class->check_action_security({ id       => $id, 
                                               required => SPOPS::Secure::SEC_LEVEL_READ });
  	}
  
	# Do any actions the class wants before fetching -- note that if any
	# of the actions returns undef (false), we bail.
  
  return undef unless ( $class->pre_fetch_action( { %{ $p }, id => $id } ) );
  
	my $obj = undef;
  
	# If we were passed the data for an object, go ahead and create it;
	# if not, check to see if we can whip up a cached object
  
	if ( $p->{data} ) {
		$obj = $class->new( $p->{data} );
  }
  else {
		$obj = $class->get_cached_object( { %{ $p }, id => $id } );
		$p->{skip_cache}++;         # Set so we don't re-cache it later
	}
  
	unless ( ref $obj eq $class ) {

		##-----  BEGIN OVERRIDING CODE  -----
		my $table_name		= $class->base_table;
		my $my_id_field		= $class->id_field;
		my $raw_fields		= [];
		my $select_fields	= [];
		my $sqltables		= [];
		my $sqlwhere		= [];
		foreach my $parent_class ( @{$class->isa_classes} ) {
			my $table		= $parent_class->table_name;
			my $id_field	= $parent_class->id_field;

			push @$sqltables, $table;	## list of tables for "FROM" clause
	
			## join tables by id field (set all equal to id field of this class)
			push @$sqlwhere,	$table . '.' . $id_field . ' = ' .
								$table_name . '.' . $my_id_field
									unless($table_name eq $table);
	
			## all fields, except id of inherited tables
			foreach my $field ( keys %{$parent_class->field} ) {
				next if $parent_class ne $class && $field eq $id_field;
				push @$select_fields, $table . '.' . $field;
				push @$raw_fields, $field;
			}
		}
		push @$sqlwhere, $class->id_clause( $id, undef, $p );

		my %args = (	from   => $sqltables,
						select => $select_fields,
						where  => join(' AND ', @$sqlwhere),
						db     => $p->{db},
						return => 'single' );

		##-----  END OVERRIDING CODE  -----

		my $row = eval { $class->db_select( \%args ); };
		
		# Keep the SQLInterface error messages in place
		
		if ( $@ ) {
			$class->fail_fetch( \%args );	 
			_w( 0, "FETCH failed. Error returned: $@" );
			die $SPOPS::Error::user_msg;
		}

		# If the row isn't found, return nothing; just as if an incorrect
		# (or nonexistent) ID were passed in

		return undef unless ( $row );
    
		# Note that we pass $p along to the ->new() method, in case
		# other information was passed in needed by it -- however, 
		# we need to be careful that certain parameters used by this method
		# (e.g., the optional 'field_alter') is not the same as a parameter
		# of an object -- THAT would be fun to debug...

		$obj = $class->new( { id => $id, %{ $p } } );
		$obj->_fetch_assign_row( $raw_fields, $row );
	}
	return $obj->_fetch_post_process( $p, $level );
}

sub fetch_group {
	my ( $class, $p ) = @_;
	my $table_name = $class->CONFIG->{base_table};
	my $p_original = { %$p };

	##-----  BEGIN OVERRIDING CODE  -----

	## At the moment this code assumes you want to fetch ALL fields
	my $my_id_field		= $class->id_field;
	my $raw_fields		= [];
	my $select_fields	= [];
	my $sqltables		= $p->{from} || [] ;
	my $sqlwhere		= [];

	foreach my $parent_class ( @{$class->isa_classes} ) {
		my $table		= $parent_class->table_name;
		my $id_field	= $parent_class->id_field;

		push @$sqltables, $table;	## list of tables for "FROM" clause

		## join tables by id field (set all equal to id field of this class)
		push @$sqlwhere,	$table . '.' . $id_field . ' = ' .
							$table_name . '.' . $my_id_field
								unless($table_name eq $table);

		## all fields, except id of inherited tables
		foreach my $field ( keys %{$parent_class->field} ) {
			next if $parent_class ne $class && $field eq $id_field;
			push @$select_fields, $table . '.' . $field;
			push @$raw_fields, $field;
		}
	}

	## original WHERE clause
	push @$sqlwhere, $p->{where}	if $p->{where};

	$p->{where}  = join(' AND ', @$sqlwhere);
	$p->{from}   = $sqltables;
	$p->{select} = $select_fields;
	$p->{return} = 'sth';
	$p->{DEBUG}  ||= SPOPS::DBI::DEBUG_FETCH;

	my $sth = $class->db_select( $p );

	##-----  END OVERRIDING CODE  -----

	# initialize limit tracking vars -- the limit passed in can be:
	# limit => 'n,n'	--> limit => 'offset,max'
	# limit => 'n'		--> limit => 'max'

	my ($count, $offset, $max) = 0;
	if ( $p->{limit} ) {
		if ( $p->{limit} =~ /,/ ) {
			( $offset, $max ) = split /\s*,\s*/, $p->{limit};
			$max += $offset;
		} 
		else {
			$max = $p->{limit};
		}
		DEBUG() && _wm( 1, SPOPS::DBI::DEBUG_FETCH, "Limit set: Start $offset to $max" );
	}

	my @obj_list = ();

ROW:
	while ( my $row = $sth->fetchrow_arrayref ) {
		##-----  MORE OVERRIDING CODE  -----
		## find the class to use for the call to fetch()
		my ($classname_idx) = grep $raw_fields->[$_] eq 'class', (0..$#{$raw_fields});
		my ($id_field_idx)  = grep $raw_fields->[$_] eq $class->id_field, (0..$#{$raw_fields});
		my $newclass = $row->[ $classname_idx ];
		my $obj = $newclass->fetch( $row->[ $id_field_idx ], $p_original);
		##-----  END MORE OVERRIDING CODE  -----

		next ROW unless ( $obj );

		## deleted security check (done by fetch)

		# Not to the offset yet, so go to the next row but still increment
		# the counter so we calculate limits properly
    
		if ( $offset and ( $count < $offset ) ) {
			$count++;
			next ROW;
		}
		last ROW if ( $max and ( $count >= $max ) );
		$count++;

		## deleted _post_fetch_process (done by fetch)

		push @obj_list, $obj;
	}
	$sth->finish;
	return \@obj_list;
}

sub isa_classes		 { return [ keys %{ $_[0]->_isa_classes } ]; }

sub inherited_fields {
	my $class	= shift;
	$class = ref($class)	if ref($class);	## get class if passed an object
	
	my $fields = [];
	foreach my $c ( @{$class->isa_classes} ) {
		next if $c eq $class;
		foreach my $field ( @{$c->field_list} ) {
			push @$fields, $field	unless $field eq $c->id_field;
		}
	}

	return $fields;
}

sub all_fields		 { return [ @{$_[0]->field_list}, @{$_[0]->inherited_fields} ]; }

sub all_field_types {
	my $class	= shift;
	my $p		= shift;
	$class = ref($class)	if ref($class);	## get class if passed an object
	
	my $type_info = {};
	foreach my $c ( @{$class->isa_classes} ) {
		my $c_types = $c->db_discover_types( $c->base_table, $p );
		foreach my $field ( @{$c->field_list} ) {
			$type_info->{$field} = $c_types->{$field}
				unless $field eq $c->id_field && $c ne $class; ## skip parent table ids
		}
	}

	return $type_info;
}

sub compare {
	my $this	= shift;
	my $that	= shift;
	my $p		= shift;

	## must be objects of the same type
	return 0 unless ref($this) eq ref($that);
	
	## and their fields must all have the same values
	my $type_info = $this->all_field_types($p);
	foreach my $field ( @{$this->all_fields} ) {
		next if $field eq $this->id_field;
		next unless defined $this->{$field} || defined $that->{$field};
		return 0 unless defined $this->{$field} && $that->{$field};
		if ( ref($this->{$field}) ) {
			return 0 unless $this->{$field}->compare($that->{$field});
		} else {
			if ( $this->_is_numeric_type($type_info->{$field}) ) {
				return 0 unless $this->{$field} == $that->{$field};
			} else {
				return 0 unless $this->{$field} eq $that->{$field};
			}
		}
	}

	return 1;
}

sub as_string {
	my $self = shift;
	my $tab = shift || '';
	my $fields = $self->CONFIG->{as_string_order} || $self->all_fields;
	my $msg = '';
	foreach my $field (@$fields){
		$msg .= $tab.sprintf( "%-20s: %s\n", $field, defined $self->{$field} ? $self->{$field} : '');
		my $ref = ref $self->{$field};
		next unless ($ref  && $ref ne 'HASH' && $ref ne 'ARRAY');
		$msg .= $self->{$field}->as_string("\t");
	}
	return $msg;
}

sub get_main_alias {
	## This method can be used to find the main alias even
	## before the class's configuration has been processed.
	my $class = shift;
	my $conf = eval '$' . $class . '::CONF';
	my ($alias) = grep $conf->{$_}->{class} eq $class, keys %$conf;

	return $alias;
}

sub get_CONF 		 { return eval '$' . $_[0] . '::CONF'; }
sub get_links_to	 { return $_[0]->get_CONF->{$_[0]->get_main_alias}->{links_to}; }

sub config_and_init {
	my $class = shift;
	
	SPOPS::ClassFactory->create( $class->_build_conf )
		unless $class->_config_processed; 

	$class->class_initialize;
}

sub _build_conf {
	my $class = shift;
	my $conf  = shift || {};

	# get the conf variable for the class.
	my $class_conf = $class->get_CONF;
	# get the alias for the class
	my $alias = $class->get_main_alias;

	unless (exists ($conf->{$alias})) {
		$conf->{$alias} = $class_conf->{$alias};
	}
	my $class_links = $class->get_links_to || '';
	if ($class_links) {
		foreach my $key (keys %$class_links) {
			next if ($key->_config_processed || 
			         exists $conf->{$key->get_main_alias});
			$conf = $key->_build_conf($conf);
		}  
	}#end if 
	return $conf;
}

sub _config_processed {
	no strict 'refs';
	my $CONFIG_method = *{$_[0]."::CONFIG"}{CODE};
	return ref($CONFIG_method) eq 'CODE';
}

sub _is_numeric_type {
	my $this		= shift;
	my $type_info	= shift;
	
	return grep $type_info == $_, (	SQL_NUMERIC,
									SQL_DECIMAL,
									SQL_INTEGER,
									SQL_SMALLINT,
									SQL_FLOAT,
									SQL_REAL,
									SQL_DOUBLE,
									SQL_BIGINT,
									SQL_TINYINT	);
}

sub _isa_classes {
	my $class	= shift;
	my $href	= shift || {};

	$class = ref($class)	if ref($class);	## get class if passed an object
	$href->{$class} = 1;					## stick it as a key in the hash
	
	foreach my $parent ( @{$class->CONFIG->{isa}} ) {
		next unless $parent->isa(ROOT_OBJ_CLASS);
		$href =  $parent->_isa_classes($href)	unless $href->{$parent};
	}
	return $href;
}

__PACKAGE__->config_and_init;

1;
__END__

=head1 NAME

ESPOPS::Object - Enhanced SPOPS Object.

=head1 SYNOPSIS

  Assuming MyObject is a subclass of ESPOPS::Object ...

  use MyObject;
  $obj = new MyObject;
  $obj = $obj->save;
  $obj = MyObject->fetch($id);
  $obj->remove;


=head1 DESCRIPTION

This is the base class for all ESPOPS (Enhanced SPOPS) objects. Where
standard SPOPS (or more correctly, SPOPS::DBI) implements persistence
for simple class of objects whose data is stored in a single table,
ESPOPS implements persistence for a set of classes with inheritance
relationships, where the attributes are distributed across several
tables, one for each class in the inheritance hierarchy.

For example, suppose we have a Vehicle class inheriting from
ESPOPS::Object, which has the fields 'name' and 'owner'. And suppose
Vehicle has a subclass Aircraft which adds the field 'ceiling'.
ESPOPS::Object also defines the fields 'id' and 'class'. In this
example, an Aircraft object will be stored into 3 tables, 'id' and
'class' in the base_table for ESPOPS::Object, 'name' and 'owner' in the
base_table for Vehicle and 'ceiling' in the base_table for Aircraft.
Each table also has an id_field which is used to join the pieces of the
object together from the 3 tables.

Also, unlike the typical usage of SPOPS objects, where the classes are
created by SPOPS and have no corresponding .pm file, ESPOPS objects are
defined in a .pm file just like a standard Perl object, with a few
additions. Each class must define the variables @ISA, $CONF, and
$TABLE_DEF in the BEGIN block. The @ISA variable is standard Perl and
$TABLE_DEF contains an SQL statement which creates the table for the
corresponding class. The $CONF variable contains an SPOPS configuration
hash with the configuration for this class only. The BEGIN block is
followed by 'use' statements for the classes which are referenced in
@ISA and the 'has_a' and 'links_to' parts of $CONF. Finally, after all
of the use statements, it should have the line:

  __PACKAGE__->config_and_init;

By convention we put it as the last line of code in the file.

These conventions allow us to say ...

  use MyObject;

... just like we would 'use' any other Perl object.

=head1 METHODS

Public Methods

=over 4

=item * $object = CLASS->new(\%)

Overrides the inherited new() method to allow the input hashref to
initialize inherited fields. Also puts the class name in the 'class' field.

=item *	$object = CLASS->fetch($id, \%)

Overrides the inherited fetch method and modifies the database SELECT.
Instead of selecting from a single table it joins the tables by id to
get all fields of the object, including inherited fields.

=item *	\@ = CLASS->fetch_group(\%)

Overrides the inherited fetch_group method and modifies the database
SELECT. Instead of selecting from a single table it joins the tables by
id to get all fields of the object, including inherited fields. This
also allows the where clause to include conditions on inherited fields.
The objects fetched are of the type specified by the 'class' field in
the Object table, not simply the class used to call fetch_group.

=item *	CLASS->config_and_init

Processes the configuration stored in $CONF for this class (via
SPOPS::Configure::DBI->process_config) and any classes related to it via
'links_to' relationships. Calls class_initialize (inherited).

=item *	\@ = CLASS->isa_classes

Returns an arrayref of all classes in this class's inheritance hierarchy
which inherit from ESPOPS::Object, including the class itself.

=item *	\@ = CLASS->inherited_fields

Returns an arrayref of all inherited non-id field names for the current
class. Does not include fields defined in the current class, only those
from parent classes.

=item *	\@ = CLASS->all_fields

Returns an arrayref of all fields for the current class, including
inherited fields.

=item *	\% = CLASS->all_field_types(\%)

Returns a hashref of SQL types for each field (including inherited
fields), where the types are those defined by DBI. Takes an optional
hashref of parameters which are passed to db_discover_types().

=item *	$ = $object1->compare($object2)

Returns 1 if the two objects contain the same data, 0 otherwise.
Note: does not compare the id fields.

=item *	$ = $object->as_string

Overrides as_string method of parent class. Prints all fields, including
inherited fields, and also prints contents of nested objects (though not
fields containing hashrefs or arrayrefs).

=item *	$ = CLASS->get_main_alias

Returns the name of the main alias for a class (the key to use in
$CONF). This method can be used even before the class's configuration
has been processed.

=item *	\% = CLASS->get_CONF

Returns the $CONF variable for the class.

=item *	\% = CLASS->get_links_to

Returns the 'links_to' configuration hashref for the class.

=back

Private Methods

=over 4

=item *	$object = $object->pre_save_action(\%)

Called autmatically when doing $object->save(). This method pre-saves the inherited fields in
the tables corresponding to parent classes, getting the ID from the top level table (for a newly
saved object). Accepts a hashref of parameters (passed in by save()).

=item *	$object = $object->pre_remove_action(\%)

Called automatically when doing $object->remove(). It removes the
corresponding row in the table corresponding to each inherited class. If
present, it also removes any corresponding rows in 'links_to' tables.

=item *	\% = CLASS->_build_conf(\%)

Used recursively to build up a hashref containing the configuration info
($CONF variable contents) for all classes linked to this class via
'links_to' definitions (actually only those whose config has not yet
been processed). All of these classes must be passed to process_config()
at the same time to properly process the configurations.

=item *	$ = CLASS->_config_processed

Returns 1 if the class's configuration has already been processed, 0
otherwise.

=item *	\% = CLASS->_isa_classes

Recursively builds a hashref whose keys contain the names of all classes
in this class's inheritance hierarchy which inherit from ESPOPS::Object,
including the class itself.

=back

=head1 AUTHORS

  Ray Zimmerman, <rz10@cornell.edu>
  Eliot Night, <enight@sporks.com>
  Raj Chandran, <raj_cmenon@hotmail.com>

=head1 SEE ALSO

SPOPS(3)
SPOPS::DBI(3)
perl(1)

=cut
